from io import TextIOWrapper

from PyVili.ContainerNode import ContainerNode
from PyVili.Node import Node
from PyVili.Types import NodeType, DataType, getDefaultValueForType, getStringValueType, getValueType
from PyVili.ViliException import (
    DataNodeInvalidDataTypeException,
    DataNodeWrongAutosetException,
    DataNodeWrongDataTypeException)


class DataNode(Node):
    def __init__(self, parent=None, id=None, dataType=None):
        if dataType is None:
            dataType = id
            id = parent
            parent = None
        super()._init(parent, id)
        if dataType is None or not isinstance(dataType, DataType):
            validType = False
            for vtype in [str, float, int, bool]:
                if type(dataType) is vtype:
                    self.data = dataType
                    self.dataType = getValueType(self.data)
                    validType = True
                    break
            if not validType:
                raise DataNodeInvalidDataTypeException(
                    {"path": self.getNodePath(), "type": dataType})
        else:
            self.dataType = dataType
            self.data = None
            self.autoset(getDefaultValueForType(self.dataType))
        self.type = NodeType.DataNode

    def checkType(self):
        if isinstance(self.data, str) and self.dataType == DataType.String:
            return True
        if (isinstance(self.data, int) or isinstance(self.data, float)) and (
                self.dataType == DataType.Int or self.dataType == DataType.Float):
            return True
        if isinstance(self.data, bool) and self.dataType == DataType.Bool:
            return True
        return False

    def set(self, data: any):
        self.data = data
        self.checkType()

    def autoset(self, data: str):
        if getStringValueType(data) == self.dataType:
            if self.dataType == DataType.Int:
                self.data = int(data)
            elif self.dataType == DataType.Float:
                self.data = float(data)
            elif self.dataType == DataType.String:
                self.data = data[1:-1]
                print("AFFECT = =================> ", data)
            elif self.dataType == DataType.Bool:
                self.data = True if data == "True" else False
            else:
                raise DataNodeWrongAutosetException(
                    {"path": self.getNodePath(), "data": data})
            self.checkType()
        else:
            raise DataNodeWrongAutosetException(
                {"path": self.getNodePath(), "data": data})

    def dumpData(self):
        print(self.id, self.data, self.dataType)
        if self.dataType == DataType.String:
            return "\"" + self.data + "\""
        elif self.dataType in [DataType.Int, DataType.Float]:
            return str(self.data)
        elif self.dataType == DataType.Bool:
            return "True" if self.data else "False"
        else:
            raise DataNodeWrongDataTypeException(
                {"path": self.getNodePath(), "datatype": self.dataType})

    def getDataType(self):
        return self.dataType

    def copy(self, newParent: ContainerNode, newId: str = ""):
        pass

    def write(self, writeFile: TextIOWrapper):
        from PyVili import ViliParser
        if self.visible:
            for i in range(self.getDepth()):
                writeFile.write(ViliParser.spacing * " ")
            writeFile.write((self.id + ":" + self.dumpData())
                            if self.id[0] != "#" else self.dumpData())
            writeFile.write("\n")



from enum import Enum

from PyVili.Utils import isStringInt, isStringFloat


class NodeType(Enum):
    Node = 0
    ContainerNode = 1
    DataNode = 2
    ArrayNode = 3
    ComplexNode = 4
    LinkNode = 5


class DataType(Enum):
    String = 0
    Int = 1
    Float = 2
    Bool = 3
    Range = 4
    Link = 5
    Template = 6
    Unknown = 7


def getDefaultValueForType(dataType: DataType):
    if dataType == DataType.String:
        return "\"\""
    elif dataType == DataType.Int:
        return "0"
    elif dataType == DataType.Float:
        return "0.0"
    elif dataType == DataType.Bool:
        return "False"
    elif dataType == DataType.Range:
        return "0..1"
    elif dataType == DataType.Link:
        return "&()"
    elif dataType == DataType.Template:
        return "T()"
    else:
        return ""


def getStringValueType(val: str):
    if val in ["True", "False"]:
        return DataType.Bool
    elif val.count("..") == 1 and len(val.split("..")) == 2:
        if isStringInt(val.split("..")[0]) and isStringInt(val.split("..")[1]):
            return DataType.Range
    elif val[0::2] == "&(" and val[-1] == ")":
        return DataType.Link
    elif val[0].isalpha() and val.count("(") == 1 and val[-1] == ")":
        return DataType.Template
    elif val[0] == '"' and val[-1] == '"' and len(val) >= 2:
        return DataType.String
    elif isStringFloat(val):
        return DataType.Float
    elif isStringInt(val):
        return DataType.Int
    else:
        return DataType.Unknown
    return None


def getValueType(val: any):
    if type(val) is int:
        return DataType.Int
    if type(val) is float:
        return DataType.Float
    if type(val) is str:
        return DataType.String
    if type(val) is bool:
        return DataType.Bool
    else:
        return DataType.Unknown



def isStringInt(s: str):
    if s[0] in ('-', '+'):
        return s[1:].isdigit()
    return s.isdigit()


def isStringFloat(s: str):
    if len(s) > 0:
        s = s[1::] if s[0] == "-" else s
        if s.count(".") == 1 and s[0] != ".":
            if isStringInt(s.replace(".", "")):
                return True
    return False



from collections import OrderedDict
from io import TextIOWrapper

from PyVili.ArrayNode import ArrayNode
from PyVili.ContainerNode import ContainerNode
from PyVili.DataNode import DataNode
from PyVili.Node import Node
from PyVili.Types import NodeType
from PyVili.ViliException import (
    ComplexNodeNoPathPartsError,
    ComplexNodeIncorrectChildTypeException,
    ComplexNodeEmptyPathException)


def convertPath(path: str) -> list:
    if "." in path:
        rpath = list(filter(lambda a: a != "", path.split(".")))
        return rpath
    elif path != "":
        return [path]
    else:
        return []


class ComplexNode(ContainerNode):
    def __init__(self, parent: ContainerNode, id: str = ""):
        super()._init(parent, id)
        self.children = OrderedDict()
        self.type = NodeType.ComplexNode

    def extractElement(self, element: Node) -> Node:
        for index, elem in enumerate(self.children):
            if elem is element:
                self.removeOwnership(elem)
                self.children.pop(elem.getId())
                return elem

    def __getitem__(self, key: str) -> Node:
        return self.children[key]

    def get(self, key: str) -> Node:
        return self.children[key]

    def getDataNode(self, key: str) -> DataNode:
        if self.get(key).getType() == NodeType.DataNode:
            return self.children[key]
        raise ComplexNodeIncorrectChildTypeException({"path": self.getNodePath(),
                                                      "type": self.get(key).getType(), "child": key})

    def getComplexNode(self, key: str) -> 'ComplexNode':
        if self.get(key).getType() == NodeType.ComplexNode:
            return self.children[key]
        raise ComplexNodeIncorrectChildTypeException({"path": self.getNodePath(),
                                                      "type": self.get(key).getType(), "child": key})

    def getArrayNode(self, key: str) -> 'ArrayNode':
        if self.get(key).getType() == NodeType.ArrayNode:
            return self.children[key]
        raise ComplexNodeIncorrectChildTypeException({"path": self.getNodePath(),
                                                      "type": self.get(key).getType(), "child": key})

    def at(self, *pathParts: str):
        if len(pathParts) >= 2:
            return self.get(pathParts[0]).at(*pathParts[1::])
        elif len(pathParts) == 1:
            return self.get(pathParts[0])
        else:
            raise ComplexNodeNoPathPartsError({"path": self.getNodePath()})

    def getPath(self, path: str) -> Node:
        sPath = convertPath(path)
        if len(sPath) > 0:
            pathIndex = 0
            getToPath = self.getComplexNode(sPath[0])
            while pathIndex != len(sPath):
                getToPath = getToPath.getComplexNode(sPath[pathIndex])
                pathIndex += 1
            return getToPath
        raise ComplexNodeEmptyPathException({"path": self.getNodePath()})

    def getAll(self, nodeType: NodeType = NodeType.Node):
        elems = []
        for elem in self.children.values():
            if elem.getType() == nodeType or nodeType == NodeType.Node:
                elems.append(elem)
            elif elem.getType in [NodeType.ComplexNode, NodeType.ArrayNode] and nodeType == NodeType.ContainerNode:
                elems.append(elem)
        return elems

    def __contains__(self, item: str):
        if item in self.children:
            return True
        return False

    def contains(self, nodeType: NodeType, key: str):
        if key in self.children and self.children[key].getType() == nodeType:
            return True
        return False

    def push(self, node: Node):
        self.children[node.getId()] = node
        node.setParent(self)
        return node

    def write(self, writeFile: TextIOWrapper):
        from PyVili import ViliParser
        if self.visible:
            for i in range(self.getDepth()):
                writeFile.write(ViliParser.spacing * " ")
            writeFile.write(self.id + ":\n")
            for child in self.children:
                self.children[child].write(writeFile)
            writeFile.write("\n")

    def copy(self, newParent: ContainerNode, newId: str = ""):
        pass



from io import TextIOWrapper

from PyVili.ViliException import (
    NodeAlreadyHaveParentException,
    NodeChangeIDWithParentException,
    NodeWrongParentRemoverException,
    NodeInvalidIdException,
    NodeInvalidParentException)

from PyVili.Types import NodeType


class Node:
    def _init(self, parent: 'ContainerNode', id: str = ""):
        from PyVili import ContainerNode
        if isinstance(parent, str) and id == "":
            self.id = parent
            self.parent = None
        else:
            self.parent = parent
            self.id = id
        self.type = NodeType.Node
        self.annotation = ""
        self.visible = True
        if not isinstance(self.id, str) or self.id == "":
            raise NodeInvalidIdException({"id": id})
        if self.parent is not None and not isinstance(self.parent, ContainerNode):
            raise NodeInvalidParentException({"parent": parent, "id": id})

    def removeParent(self, currentParent):
        if self.parent is None or self.parent == currentParent:
            self.parent = None
        else:
            raise NodeWrongParentRemoverException(
                {"wrongparent": currentParent.getNodePath(), "path": self.getNodePath()}
            )

    def setAnnotation(self, annotation: str):
        self.annotation = annotation

    def getAnnotation(self):
        return self.annotation

    def getId(self):
        return self.id

    def setId(self, id: str):
        if self.parent is None:
            self.id = id
        else:
            raise NodeChangeIDWithParentException(
                {"path": self.getNodePath(), "parent": self.parent.getNodePath()}
            )

    def getType(self):
        return self.type

    def setParent(self, parent: 'ContainerNode'):
        if self.parent is None:
            self.parent = parent
        else:
            raise NodeAlreadyHaveParentException(
                {"path": self.getNodePath(), "parent": self.parent.getNodePath()}
            )

    def getParent(self):
        return self.parent

    def getNodePath(self):
        parentChain = []
        currentParent = self.parent
        while currentParent is not None:
            parentAnnotation = "" if currentParent.getAnnotation() == "" else "<" + \
                currentParent.getAnnotation() + ">"
            parentChain.append(currentParent.getId() + parentAnnotation)
            currentParent = currentParent.getParent()
        parentChain = parentChain[::-1]
        parentChain.append(self.id + "" if self.annotation
                           == "" else "<" + self.annotation + ">")
        return "/".join(parentChain)

    def getRawNodePath(self):
        parentChain = []
        currentParent = self.parent
        while currentParent is not None:
            parentChain.append(currentParent.getId())
            currentParent = currentParent.getParent()
        parentChain = parentChain[::-1]
        parentChain.append(self.id)
        return "/".join(parentChain)

    def getDepth(self):
        currentParent = self.parent
        depth = 0
        while currentParent is not None:
            currentParent = currentParent.getParent()
            depth += 1
        return depth

    def getVisible(self):
        return self.visible

    def setVisible(self, visible: bool):
        self.visible = visible

    def copy(self, newParent: 'ContainerNode', newId: str = ""):
        raise NotImplementedError("Node::copy is abstract method")

    def write(self, writeFile: TextIOWrapper):
        raise NotImplementedError("Node::write is abstract method")



from io import TextIOWrapper

from PyVili.ContainerNode import ContainerNode
from PyVili.DataNode import DataNode
from PyVili.Node import Node
from PyVili.Types import getValueType, NodeType


class ArrayNode(ContainerNode):
    def __init__(self, parent: ContainerNode, id: str = ""):
        super()._init(parent, id)
        self.children = []
        self.iterIndex = 0
        self.type = NodeType.ArrayNode

    def reorder(self, index: int):
        for i in range(index, len(self.children)):
            self.removeOwnership(self.children[i])
            self.children[i].setId("#" + str(i))
            self.children[i].setParent(self)

    def size(self):
        return len(self.children)

    def get(self, index: int):
        return self.children[index]

    def __getitem__(self, index: int):
        return self.children[index]

    def __iter__(self):
        return self

    def __next__(self):
        if self.iterIndex >= self.size():
            self.iterIndex = 0
            raise StopIteration
        else:
            self.iterIndex += 1
            return self.children[self.iterIndex - 1]

    def push(self, value: any):
        newNode = DataNode(
            self, "#" + str(len(self.children)), getValueType(value))
        newNode.set(value)
        self.children.append(newNode)

    def insert(self, index: int, value: any):
        newNode = DataNode(self, "#" + str(index), getValueType(value))
        newNode.set(value)
        self.children.insert(0, newNode)

    def clear(self):
        self.children.clear()

    def erase(self, index: int):
        self.children.remove(index)

    def extractElement(self, element: Node):
        for index, elem in enumerate(self.children):
            if elem is element:
                self.removeOwnership(elem)
                self.erase(index)

    def copy(self, newParent: ContainerNode, newId: str = ""):
        pass

    def write(self, writeFile: TextIOWrapper):
        from PyVili import ViliParser
        if self.visible:
            for i in range(self.getDepth()):
                writeFile.write(ViliParser.spacing * " ")
            writeFile.write(self.id + ":[\n")
            for child in self.children:
                child.write(writeFile)
            for i in range(self.getDepth()):
                writeFile.write(ViliParser.spacing * " ")
            writeFile.write("]\n")



"""Manipulate vili files in Python
"""

from setuptools import setup, find_packages
from os import path

from vili import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='vili',
    version=__version__,
    description='PyVili is a Python API to manipulate Vili Data files',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Sygmei/PyVili',
    author='Sygmei',
    author_email='sygmei@sygmei.io',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='vili data language config',
    packages=find_packages(include=['vili']),
    python_requires='>=3.6, <4',
    install_requires=[],
    extras_require={
        'test': [
            'coverage',
            'tox',
            'pytest',
            'flake8'
        ],
    },
    entry_points={
        'console_scripts': [],
    },
    project_urls={
        'Bug Reports': 'https://github.com/Sygmei/PyVili/issues',
        'Source': 'https://github.com/Sygmei/PyVili',
    },
)



from io import TextIOWrapper

from PyVili.Node import Node
from PyVili.Types import NodeType


class ContainerNode(Node):
    def _init(self, parent: 'ContainerNode', id: str = ""):
        super()._init(parent, id)
        self.type = NodeType.ContainerNode

    def removeOwnership(self, element: Node):
        element.removeParent(self)
        return element

    def extractElement(self, element: Node):
        raise NotImplementedError(
            "ContainerNode::extractElement is abstract method")

    def copy(self, newParent: 'ContainerNode', newId: str = ""):
        raise NotImplementedError("ContainerNode::copy is abstract method")

    def write(self, writeFile: TextIOWrapper):
        raise NotImplementedError("ContainerNode::write is abstract method")



from PyVili.ArrayNode import *
from PyVili.ComplexNode import *
from PyVili.ContainerNode import *
from PyVili.DataNode import *
from PyVili.Node import *
from PyVili.Types import *
from PyVili.Utils import *
from PyVili.ViliException import *
from PyVili.ViliParser import *

__all__ = ["ArrayNode", "ComplexNode", "ContainerNode", "DataNode", "Node", "Types", "Utils", "ViliException", "ViliParser"]
__version__ = "1.1.0"


class ViliException(Exception):
    def __init__(self, msg, dct):
        self.msg = msg
        for word in dct:
            self.msg = self.msg.replace("%" + word + "%", str(dct[word]))
        super().__init__(self.msg)


class NodeAlreadyHaveParentException(ViliException):
    def __init__(self, dct):
        super().__init__("Node %path% already has a parent : %parent%", dct)


class NodeChangeIDWithParentException(ViliException):
    def __init__(self, dct):
        super().__init__("Can't change id of %path% when it has a parent : %parent%", dct)


class NodeWrongParentRemoverException(ViliException):
    def __init__(self, dct):
        super().__init__("%wrongparent% is not %path% parent", dct)


class NodeInvalidIdException(ViliException):
    def __init__(self, dct):
        super().__init__("Invalid Id for Node : '%id%'", dct)


class DataNodeInvalidDataTypeException(ViliException):
    def __init__(self, dct):
        super().__init__("Invalid DataType : '%type%' for DataNode : %path%", dct)


class NodeInvalidParentException(ViliException):
    def __init__(self, dct):
        super().__init__("Invalid Parent : '%parent%' for Node : %id%", dct)


class DataNodeWrongAutosetException(ViliException):
    def __init__(self, dct):
        super().__init__("Can't autoset raw value %data% for DataNode %path%", dct)


class DataNodeWrongDataTypeException(ViliException):
    def __init__(self, dct):
        super().__init__("DataType %datatype% is not valid for DataNode %path%", dct)


class ComplexNodeNoPathPartsError(ViliException):
    def __init__(self, dct):
        super().__init__("ComplexNode %path% <at> method call must have at least one path part", dct)


class ComplexNodeIncorrectChildTypeException(ViliException):
    def __init__(self, dct):
        super().__init__("Child '%child%' of ComplexNode '%path%' has type %type%", dct)


class ComplexNodeEmptyPathException(ViliException):
    def __init__(self, dct):
        super().__init__("ComplexNode '%path%' can't access empty path", dct)
